import numpy as np
import sys
import os
import h5py
import re
import xarray as xr #cite as: Hoyer, S. & Hamman, J., (2017). xarray: N-D labeled Arrays and Datasets in Python. Journal of Open Research Software. 5(1), p.10. DOI: http://doi.org/10.5334/jors.148
import gc
import glob
import copy
#from sklearn.neighbors import BallTree
import matplotlib.pyplot as plt
import netCDF4 as nv
from scipy.interpolate import interp1d,interp2d,griddata
from scipy.interpolate import RectBivariateSpline as rbs
from scipy.interpolate import RegularGridInterpolator as rgi
from scipy.optimize import brentq, curve_fit, minimize_scalar, minimize, Bounds
from scipy.stats import binned_statistic_2d, binned_statistic_dd, binned_statistic
import pickle
import cartopy.crs as ccrs
from matplotlib import rcParams

rcParams['ps.useafm'] = True
rcParams['pdf.use14corefonts'] = False
rcParams['text.usetex'] = False
rcParams['font.sans-serif'] = ['cmr10', 'Times-Roman']
rcParams['font.weight'] = 'normal'
rcParams['font.size'] = 20
rcParams['xtick.labelsize'] = 20
rcParams['ytick.labelsize'] = 20
rcParams['axes.labelsize'] = 22
rcParams['axes.linewidth'] = 1
rcParams['axes.unicode_minus'] = False
rcParams['xtick.minor.width'] = 1
rcParams['ytick.minor.width'] = 1
rcParams['xtick.major.width'] = 1
rcParams['ytick.major.width'] = 1
rcParams['xtick.minor.size'] = 4
rcParams['ytick.minor.size'] = 4
rcParams['xtick.major.size'] = 5
rcParams['ytick.major.size'] = 5
rcParams['xtick.direction'] = 'in'
rcParams['ytick.direction'] = 'in'
rcParams['mathtext.fontset'] = 'cm'


import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.colors as colors
from matplotlib import gridspec
from matplotlib import ticker
import matplotlib as mpl
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.ticker import NullFormatter
from collections import OrderedDict, Counter
from matplotlib.lines import Line2D
from matplotlib.patches import Patch, Rectangle, Ellipse
import matplotlib.patheffects as pe

from math import radians, cos, sin, asin, sqrt


def haversine(lon1,lat1,lon2,lat2):
  # convert decimal degrees to radians 
  lo1 = np.radians(lon1)
  lo2 = np.radians(lon2)
  la1 = np.radians(lat1)
  la2 = np.radians(lat2)

  # haversine formula 
  dlon = lo2 - lo1 
  dlat = la2 - la1 
  a = np.sin(dlat/2)**2 + np.cos(la1) * np.cos(la2) * np.sin(dlon/2)**2
  c = 2. * np.arcsin(np.sqrt(a)) 
  r = 6371. # Radius of earth in km.
  return c * r
"""
------------------------------------------

------------------------------------------

-----------Notes--------------------------------------------------------
"""

class PartInfo():
  """
  Fields in partposit files
  ---------------------------------
  xlon = latitude (degrees)
  ylat = longitude (degrees)
  z = heigth (meters)
  topo = topography
  tro = tropopause
  hmix = PBL height
  rho = density
  qv = specific humidity
  pv = potential vorticity
  tt = temperature
  uu = 
  vv =
  """ 

  def __init__(self, path_to_directory="/home/lucie/SizeTesting/output_1e6_23", 
    partposit_nc_fname = "partposit_average.nc",interpolatetopo=False):

    self.path = path_to_directory
    #Read in the header
    #self.header = read_header_bin(path_to_directory=self.path)
    #Npart from the header is not accurate when domainfilling is used, so this will be set when data is read instead
    self.npart = None

    #Listing all fields present in the binary partposit files
    xmass_dtype = [('xmass_%d' % (i+1), 'f4') for i in range(1)]#self.header['nspec'])]
    self.fields_partposit = ([('npoint', 'i4'), ('xtra1', 'f4'), ('ytra1', 'f4'), ('ztra1', 'f4'),
      ('itramem', 'i4'), ('topo', 'f4'), ('pvi', 'f4'), ('qvi', 'f4'),
      ('rhoi', 'f4'), ('hmixi', 'f4'), ('tri', 'f4'), ('tti', 'f4')] + xmass_dtype)

    self.partposit_nc_fname = partposit_nc_fname
    
    #Open netcdf particledata using xarray and naming it self._pd
    self._read_partposit_nc()
    self.coords = None

    self.kdtree = {}

    #2D and 3D fields interpolation (if required)
    self.topography = None
    self.tropopause = None
    self.hmixing = None
    self.interpolatetopo = interpolatetopo

  def _read_partposit_nc(self):
    tempfile = self.path + '/' + self.partposit_nc_fname
    self._pd = xr.open_dataset(tempfile, engine="netcdf4")
    print("netCDF4 file opened.")

  @property
  def xlon(self):
    return self._pd.longitude

  @property
  def ylat(self):
    return self._pd.latitude

  @property
  def z(self):
    return self._pd.height

  @property
  def topo(self):
    if self.interpolatetopo:
      if self.topography is None:
        topo_inter = self.interpolate2d(self._pd.topo,0)
        self.topography = np.zeros((len(self.xlon),len(self.xlon[0,:])))
        for itime in range(len(self.topography[0,:])):
          topo_inter = self.interpolate2d(self._pd.topo,0)
          self.topography[:,itime]  = topo_inter(self.ylat[:,itime],self.xlon[:,itime],grid=False)
          mini=np.min(self._pd.topo).values
          maxi=np.max(self._pd.topo).values
          self.topography[self.topography[:,itime]<mini,itime]=mini
          self.topography[self.topography[:,itime]>maxi,itime]=maxi
      return self.topography
    else:
      return self._pd.topo

  @property
  def tro(self):
    if len(self._pd.tr.dims)==3:
      if self.tropopause is None:
        self.tropopause = np.zeros((len(self.xlon),len(self.xlon[0,:])))
        for itime in range(len(self.tropopause[0,:])):
          tro_inter = self.interpolate2d(self._pd.tr,itime)
          self.tropopause[:,itime] = tro_inter(self.ylat[:,itime],self.xlon[:,itime],grid=False)
          mini=np.min(self._pd.tr).values
          maxi=np.max(self._pd.tr).values
          self.tropopause[self.tropopause[:,itime]<mini,itime]=mini
          self.tropopause[self.tropopause[:,itime]>maxi,itime]=maxi
      return self.tropopause
    else:
      return self._pd.tr

  @property
  def hmix(self):
    if len(self._pd.hmix.dims)==3:
      if self.hmixing is None:
        self.hmixing = np.zeros((len(self.xlon),len(self.xlon[0,:])))
        for itime in range(len(self.hmixing[0,:])):
          hmix_inter = self.interpolate2d(self._pd.hmix,itime)
          self.hmixing[:,itime] = hmix_inter(self.ylat[:,itime],self.xlon[:,itime],grid=False)
          mini=np.min(self._pd.hmix).values
          maxi=np.max(self._pd.hmix).values
          self.hmixing[self.hmixing[:,itime]<mini,itime]=mini
          self.hmixing[self.hmixing[:,itime]>maxi,itime]=maxi
      return self.hmixing
    else:
      return self._pd.hmix

  @property
  def rho(self):
    return self._pd.rho

  @property
  def qv(self):
    return self._pd.qv

  @property
  def pv(self):
    return self._pd.pv

  @property
  def tt(self):
    if ('tt' in self._pd):
      return self._pd.tt
    else:
      return self._pd.temperature

  @property
  def mass(self):
    return self._pd.mass001

  #Only FLEXTRA
  @property
  def rr(self):
    return self._pd.rho
  #Only FLEXTRA
  @property
  def theta(self):
    return self._pd.theta
  #Only FLEXTRA
  @property
  def pressure(self):
    return self._pd.pr

  @property
  def time(self):
    return np.array(self._pd.time)

  @property
  def particle(self):
    return np.array(self._pd.particle)

  @property 
  def day(self):
    return self._pd.time.dt.day.values

  @property 
  def month(self):
    return self._pd.time.dt.month.values

  @property 
  def year(self):
    return self._pd.time.dt.year.values

  @property
  def period(self):
    return self._pd.attrs[list(self._pd.attrs.keys())[0]]

  def interpolate2d(self,field,itime):
    #lon = field.lon.values
    lon=np.arange(0.5,361,0.5)
    #lat = field.lat.values-90.
    lat=np.arange(-89.5,91,0.5)
    # x,y=np.meshgrid(lon,lat)
    # xy=np.stack([x.ravel(),y.ravel()],-1)
    # info=np.zeros_like(x)
    # if len(field.dims)==3:
    #   infotemp = field[:,:,itime].values
    #   info[1:,1:]=infotemp
    #   info[1:,0]=infotemp[:,1]
    #   info[0,1:]=infotemp[1,:]
    #   info[0,0]=infotemp[1,1]
    # else:
    #   infotemp = field[:,:].values
    #   info[1:,1:]=infotemp
    #   info[1:,0]=infotemp[:,1]
    #   info[0,1:]=infotemp[1,:]
    #   info[0,0]=infotemp[1,1]
    # return griddata((x.ravel(),y.ravel()),info.ravel(),(self.xlon[:,itime],self.ylat[:,itime]))
    if len(field.dims)==2:
      info = field[:,:].values
    else:
      info = field[:,:,itime].values
    return rbs(lat,lon,info)

def compute_conservation_properties(
  length=48,
  basepath='',
  outputpath='',
  basename='pvtest',
  dirs=['orig','eta','meter'],
  interptopo=[True,True,True],
  one_release=False,
  zrange=[1000,5000],
  ldirect=[1,1]):

  ATCE_q={}
  RTCE_q={}
  ATCE_t={}
  RTCE_t={}
  RTCE_p={}
  ATCE_p={}
  ATCE_e={}
  RTCE_e={}
  Npart={}
  zpart={}

  for i in range(len(dirs)):
    ATCE_q[i],RTCE_q[i],ATCE_t[i],RTCE_t[i],ATCE_p[i],RTCE_p[i],ATCE_e[i],RTCE_e[i],Npart[i],temp1,temp2,zpart[i] = conservation_calculation(
      path_to_directory=basepath+dirs[i],one_release=one_release,interptopo=interptopo[i],
      partposit_nc_fname='',length=length,flextra=False,maxpart=None,lower=zrange[0],upper=zrange[1],ldirect=ldirect[i])
    data_all={}
    data_all['ATCE_q']=ATCE_q[i]
    data_all['RTCE_q']=RTCE_q[i]
    data_all['ATCE_t']=ATCE_t[i]
    data_all['RTCE_t']=RTCE_t[i]
    data_all['ATCE_p']=ATCE_p[i]
    data_all['RTCE_p']=RTCE_p[i]
    data_all['ATCE_e']=ATCE_e[i]
    data_all['RTCE_e']=RTCE_e[i]
    data_all['Npart']=Npart[i]
    data_all['zpart']=zpart[i]
    datafile = open(dirs[i]+'data%i_600.pickle'%zrange[0],'wb')
    pickle.dump(data_all,datafile)
    datafile.close()



def conservation_calculation(path_to_directory="/home/lucie/",
  partposit_nc_fname="conservation_sea_0.nc",length=144,maxpart=None,
  flextra=False, press=True,one_release=False,errormap=False,interptopo=True,
  lower=1000.,upper=5000.,ldirect=1):

  pd = {}

  files = os.listdir(path_to_directory)
  fname='partoutput'
  ff = []
  times = []
  for file in files:
    if file.endswith(".nc") and file.startswith(fname):
      ff.append(file)
      times.append(int(re.findall(r'\d{14}',file)[0][0:10]))
  times = np.array(times)
  order = np.argsort(times)
  if ldirect==-1:
    ff = list(np.array(ff)[order[::-1]])
  else:
    ff = list(np.array(ff)[order])
  for file in ff:
    print(file)
    pd[file] = PartInfo(path_to_directory=path_to_directory,partposit_nc_fname=file,interpolatetopo=interptopo)
    if interptopo:
      pd[file]._pd['topo'] = pd[ff[0]]._pd['topo']
  tot = -1
  maxlength=length
  ii=0
  ntime=0
  for file in ff:
    lsum = maxlength - ntime
    if lsum <= 0:
      break

    print('Reading', file, lsum, maxlength)
    lon_temp=pd[file].xlon[:tot,:lsum].values
    lon_temp = np.where(lon_temp>180., lon_temp-360. ,lon_temp)
    lat_temp=pd[file].ylat[:tot,:lsum].values
    z_temp=pd[file].z[:tot,:lsum].values
    qv_temp=pd[file].qv[:tot,:lsum].values
    pv_temp=pd[file].pv[:tot,:lsum].values
    v_temp=pd[file].v[:tot,:lsum].values
    u_temp=pd[file].u[:tot,:lsum].values
    w_temp=pd[file].w[:tot,:lsum].values
    if press:
      p_temp=pd[file].pressure[:tot,:lsum].values
    if interptopo:#True:#len(pd[file]._pd.tr.dims)==3:
      topo_temp=pd[file].topo[:tot,:lsum]
      #tr_temp=pd[file].tro[:tot,:lsum]
    else:
      topo_temp=pd[file].topo[:tot,:lsum].values
    if flextra:
      theta_temp=pd[file].theta[:tot,:lsum].values
    else:
      tt_temp=pd[file].tt[:tot,:lsum].values

    if ii==0:
      lon = np.zeros((len(lat_temp),maxlength))
      lat = np.zeros((len(lat_temp),maxlength))
      z = np.zeros((len(lat_temp),maxlength))
      qv = np.zeros((len(lat_temp),maxlength))
      pv = np.zeros((len(lat_temp),maxlength))
      e = np.zeros((len(lat_temp),maxlength))
      if press:
        p = np.zeros((len(lat_temp),maxlength))
      topo = np.zeros((len(lat_temp),maxlength))
      #tr = np.zeros((len(lat_temp),maxlength))
      if flextra:
        theta = np.zeros((len(lat_temp),maxlength))
      else:
        tt = np.zeros((len(lat_temp),maxlength))
      ii=1


    npart_n = len(lat_temp)
    ntime_n = len(lat_temp[0])

    e[:npart_n,ntime:ntime+ntime_n]=tt_temp*1.005 + (z_temp+topo_temp)*9.807+qv_temp*2257.06+(u_temp*u_temp+v_temp*v_temp+w_temp*w_temp)/2.
    lat[:npart_n,ntime:ntime+ntime_n]=lat_temp
    lon[:npart_n,ntime:ntime+ntime_n]=lon_temp
    z[:npart_n,ntime:ntime+ntime_n]=z_temp
    qv[:npart_n,ntime:ntime+ntime_n]=qv_temp
    pv[:npart_n,ntime:ntime+ntime_n]=pv_temp
    if press:
      p[:npart_n,ntime:ntime+ntime_n]=p_temp
    topo[:npart_n,ntime:ntime+ntime_n]=topo_temp
    #tr[:npart_n,ntime:ntime+ntime_n]=tr_temp
    if flextra:
      theta[:npart_n,ntime:ntime+ntime_n]=theta_temp
    else:
      tt[:npart_n,ntime:ntime+ntime_n]=tt_temp

    ntime = ntime + ntime_n

  if flextra:
    p=p*100.
    tt=theta/((100000./p)**(0.286))
    rh=0.263*p*qv*(np.exp((17.67*(tt-273.15))/(tt-29.65)))**(-1)
  else:
    #https://calculator.academy/atmospheric-pressure-calculator/
    if not press:
      p=101325.*np.exp(-9.8*(z+topo)*0.028969/tt/8.3144626)
    #https://earthscience.stackexchange.com/questions/2360/how-do-i-convert-specific-humidity-to-relative-humidity
    rh = 0.263*p*qv*(np.exp((17.67*(tt-273.15))/(tt-29.65)))**(-1)
    theta = tt*(100000./p)**(0.286)
  del p
  #humidity <90%
  #z>1000m
  start=np.zeros(len(z)).astype(int)
  # end=np.zeros(len(z)).astype(int)
  end=np.ones(len(z)).astype(int)*length

  print('Selecting particles')
  for i in range(len(z)):
    zt=z[i][start[i]:]+topo[i][start[i]:]
    if one_release:
      if ((len(np.where((zt<lower)|(zt>upper)|(rh[i][start[i]:]>90.))[0]) > 0) or
        (np.isnan(np.sum(zt))==True) or (np.isnan(np.sum(qv[i][start[i]:]))==True)):
        start[i]=-1
    elif ((len(np.where((zt<lower)|(zt>upper)|(rh[i][start[i]:]>90.))[0]) > 0) or
      (np.isnan(np.sum(zt))==True) or (np.isnan(np.sum(qv[i][start[i]:]))==True) or
      (np.isnan(np.sum(theta[i][start[i]:]))==True) or (np.isnan(np.sum(pv[i][start[i]:]))==True) or
      (len(np.where((np.abs(lat[i][start[i]:])<40)|(np.abs(lat[i][start[i]:])>80))[0])>0)):
      start[i]=-1
  deeltjes = np.where(start==0)[0]
  if maxpart is not None and maxpart<len(deeltjes):
    deeltjes = np.random.choice(deeltjes, size=maxpart, replace=False)#deeltjes[:maxpart]#np.random.choice(deeltjes, size=maxpart, replace=False)
  del rh
  print('Computing values')
  qv = qv[deeltjes]
  theta = theta[deeltjes]
  pv=np.abs(pv[deeltjes])
  e=e[deeltjes]

  ATCE_q = np.zeros(length)
  RTCE_q = np.zeros(length)
  N = len(deeltjes)
  print(len(qv))
  ATCE_q=1./N*np.sum(np.abs(qv-np.array([qv[:,0]]).T),axis=0)
  RTCE_q=np.sum(np.abs(qv-np.array([qv[:,0]]).T),axis=0)/np.sum(np.abs(qv+np.array([qv[:,0]]).T)/2.,axis=0)*100.

  ATCE_t=1./N*np.sum(np.abs(theta-np.array([theta[:,0]]).T),axis=0)
  RTCE_t=np.sum(np.abs(theta-np.array([theta[:,0]]).T),axis=0)/np.sum(np.abs(theta+np.array([theta[:,0]]).T)/2.,axis=0)*100.

  if flextra:
    del theta,qv
  else:
    del tt,qv

  pv_diff = pv-np.array([pv[:,0]]).T
  pv_diff[np.isnan(pv_diff)]=0
  pv_diff[np.isinf(pv_diff)]=0
  ATCE_p=1./N*np.sum(np.abs(pv_diff),axis=0)
  RTCE_p=np.sum(np.abs(pv_diff),axis=0)/np.sum(np.abs(pv+np.array([pv[:,0]]).T)/2.,axis=0)*100.

  e_diff = e-np.array([e[:,0]]).T
  e_diff[np.isnan(e_diff)]=0
  e_diff[np.isinf(e_diff)]=0
  ATCE_e=1./N*np.sum(np.abs(e_diff),axis=0)
  RTCE_e=np.sum(np.abs(e_diff),axis=0)/np.sum(np.abs(e+np.array([e[:,0]]).T)/2.,axis=0)*100.


  ml = len(pv[0])
  mp = len(pv)
  if not errormap:
    del pv
    return ATCE_q,RTCE_q,ATCE_t,RTCE_t,ATCE_p,RTCE_p,ATCE_e,RTCE_e,N,ml,mp,z[deeltjes]

  grid_xlon = np.arange(-180, 180, 1.)
  grid_ylat = np.array([-90])
  i=0
  while grid_ylat[i] <= 90:
    grid_ylat = np.append(grid_ylat, grid_ylat[i] + np.abs(grid_ylat[i])/90*3.47+1)#3.47+1)#3)
    i += 1
  if grid_ylat[-1]>90:
    grid_ylat[-1] = 90.

  lon2d, lat2d = np.meshgrid((grid_xlon[1:]+grid_xlon[:-1])/2.,
    (grid_ylat[1:]+grid_ylat[:-1])/2.)
  error = np.zeros((len(grid_ylat)-1, len(grid_xlon)-1))
  aantal = np.zeros((len(grid_ylat)-1, len(grid_xlon)-1))

  lat=lat[deeltjes]
  lon=lon[deeltjes]

  for i in range(length):
    error_temp, edges, bn = binned_statistic_dd([lat[:,i],lon[:,i]], pv_diff[:,i],
      statistic='sum',bins=[grid_ylat,grid_xlon])
    aantal_temp, edges = np.histogramdd([lat[:,i],lon[:,i]], bins=[grid_ylat,grid_xlon])

    error_temp[np.isnan(error_temp)] = 0
    error = error + error_temp
    aantal = aantal + aantal_temp

  em = error/aantal
  em[np.isnan(em)]=0
  em[np.isinf(em)]=0
  return ATCE_q,RTCE_q,ATCE_t,RTCE_t,ATCE_p,RTCE_p,ATCE_e,RTCE_e,N,ml,mp,z[deeltjes],em,lon2d,lat2d

