import numpy as np
import sys
import os
import h5py
import re
import xarray as xr #cite as: Hoyer, S. & Hamman, J., (2017). xarray: N-D labeled Arrays and Datasets in Python. Journal of Open Research Software. 5(1), p.10. DOI: http://doi.org/10.5334/jors.148
import gc
import glob
import copy
import argparse
#from sklearn.neighbors import BallTree
import matplotlib.pyplot as plt
import netCDF4 as nv

import cartopy.crs as ccrs
import cartopy.feature as cf
from matplotlib import rcParams

rcParams['ps.useafm'] = True
rcParams['pdf.use14corefonts'] = False
rcParams['text.usetex'] = False
rcParams['font.sans-serif'] = ['cmr10', 'Times-Roman']
rcParams['font.weight'] = 'normal'
rcParams['font.size'] = 28
rcParams['xtick.labelsize'] = 22
rcParams['ytick.labelsize'] = 22
rcParams['axes.labelsize'] = 22
rcParams['axes.linewidth'] = 1
rcParams['axes.unicode_minus'] = False
rcParams['xtick.minor.width'] = 1
rcParams['ytick.minor.width'] = 1
rcParams['xtick.major.width'] = 1
rcParams['ytick.major.width'] = 1
rcParams['xtick.minor.size'] = 4
rcParams['ytick.minor.size'] = 4
rcParams['xtick.major.size'] = 5
rcParams['ytick.major.size'] = 5
rcParams['xtick.direction'] = 'in'
rcParams['ytick.direction'] = 'in'
rcParams['mathtext.fontset'] = 'cm'


import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.colors as colors
from matplotlib.colors import LogNorm
from matplotlib import gridspec
from matplotlib import ticker
import matplotlib as mpl
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.ticker import NullFormatter
from collections import OrderedDict, Counter
from matplotlib.lines import Line2D
from matplotlib.patches import Patch, Rectangle, Ellipse
from scipy import stats
import matplotlib.patheffects as pe

import pandas as pd


class Grid():

  def __init__(self, path_to_directory="/home/lucie/ETEX_PS_2000/output/", 
    fname = "grid_conc_19941023160000.nc"):

    self.path = path_to_directory
    self.fname = fname

    self._read_file()

  def _read_file(self):
    tempfile = self.path + '/' + self.fname
    self._data = nv.Dataset(tempfile, "r", format="NETCDF4")
    print("netCDF4 file opened.")

  @property
  def wd(self):
    return self._data['WD_spec001'][0,0,:,:,:]

  @property
  def dd(self):
    return self._data['DD_spec001'][0,0,:,:,:]

  @property
  def dataset(self):
    return self._data['spec001_mr']

  @property
  def receptordata(self):
    return self._data['receptor_conc001']

  @property
  def longitude(self):
    return np.array(self._data['longitude'][:])

  @property
  def latitude(self):
    return np.array(self._data['latitude'][:])

  @property
  def height(self):
    return np.array(self._data['height'][:])

  @property
  def receptor(self):
    return np.array(self._data['receptor'][:])

  @property
  def data(self):
    return self.dataset[0,0,:,:,:,:]

  def concentration(self,lat,lon,height=None):
    ilon=np.abs(self.longitude-lon).argmin()
    ilat=np.abs(self.latitude-lat).argmin()
    if height is None:
      return np.sum(np.array(self.data[:,0:1,ilat,ilon]),axis=1)
    else:
      iheight=np.abs(self.height-height).argmin()
      dh1=np.abs(self.height[iheight]-height)/(self.height[iheight+1]-self.height[iheight])
      return ((1.-dh1)*(self.data[:,iheight,ilat,ilon])+
        dh1*np.array(self.data[:,iheight+1,ilat,ilon]))


  def concentration_rec(self,index):
    #index=np.where(self.receptor==station)[0]
    return self.receptordata[index,:]

class Etex():

  def __init__(self, path_to_directory="/home/lucie/ETEX_PS_2000/Statistik_ASt/", 
    fname = "pmch2.dat",locname="stationlist.950130"):

    self.path = path_to_directory
    self.fname = fname
    self.locname = locname

    self._df = pd.read_csv(self.path+"/"+self.fname,delim_whitespace=True)

    self._loc = pd.read_csv(self.path+"/"+self.locname,sep=',')

  @property
  def df(self):
    return self._df
  
  @property
  def loc(self):
    return self._loc[['cc', ' Lat', ' Long', ' Alt']]

  @property
  def stations(self):
    return self.loc['cc'].to_numpy()

  def location(self,station):
    index=np.where(self.loc['cc']==station)[0][0]
    loctemp = self.loc.loc[[index]].to_numpy()[0,1:].astype(float)
    d0 = int(loctemp[0]) + (loctemp[0]-int(loctemp[0]))/.6
    d1 = int(loctemp[1]) + (loctemp[1]-int(loctemp[1]))/.6
    return np.array([d0,d1,loctemp[2]])#, loctemp

  def concentration(self,station):
    index=np.where(self.df['station']==station)[0][0]
    return self.df.loc[[index]].to_numpy()[0,1:].astype(float)


def compute_etex_concentrations(grid,etex,receptor=False):
  lengte=np.min([len(etex.concentration(etex.stations[0])),len(grid.data)])
  econc=np.zeros((len(etex.stations),lengte))
  gconc=np.zeros((len(etex.stations),lengte))
  rconc=np.zeros((len(etex.stations),lengte))

  i=0
  for station in etex.stations:
    econc_tmp=etex.concentration(station)[:len(econc[0])]
    if receptor:
      rconc_tmp=grid.concentration_rec(i)[:len(econc[0])]
      rconc_tmp[rconc_tmp<0.01]=0.
      rconc[i,:]=rconc_tmp
    ww=np.where(econc_tmp<0)[0]
    econc_tmp[(econc_tmp<0.01)]=0.0
    econc_tmp[ww]=-1
    econc[i,:]=econc_tmp
    loc=etex.location(station)
    gconc_tmp=grid.concentration(loc[0],loc[1])[:len(econc_tmp)]
    gconc_tmp[(gconc_tmp<0.01) & (gconc_tmp>=0.005)]=0.01
    gconc_tmp[(gconc_tmp<0.01)]=0.0

    gconc[i,:]=gconc_tmp
    i+=1

  return econc,gconc,rconc

def plot_concentration_map(path_to_directory=None,fname="grid_conc_19941023160000_nest.nc",path_to_etex_directory='./'):#grid,econc,gconc,tracer,fbasename='',captex=False):

  grid=Grid(path_to_directory="/home/lucie/ETEX_new/output_eta_820_2/",
    fname = "grid_conc_19941023160000_nest.nc")
  tracer=Etex(path_to_directory="/home/lucie/ETEX_PS_2000/Statistik_ASt/")
  econc,gconc,rconc=compute_etex_concentrations(grid,tracer,receptor=False)
  captex=False

  lon=grid.longitude
  lat=grid.latitude
  lonnew,latnew = np.meshgrid( lon, lat)
  s=180
  levels=np.logspace(np.log10(0.005),1,100)
  cmap=copy.copy(plt.cm.PuBu)
  colour='indigo'
  colour2='darkslategrey'
  if captex:
    totlen=8
  else:
    totlen=len(gconc[0])
  for j in range(totlen):
    fig=plt.figure(figsize=[13,9])
    ax = plt.subplot(111, projection=ccrs.PlateCarree(central_longitude=0))

    ax.add_feature(cf.COASTLINE)
    ax.add_feature(cf.BORDERS)
    ax.add_feature(cf.RIVERS)
    ax.add_feature(cf.STATES,linewidth=0.5)

    for i in range(len(tracer.stations)):
      loc=tracer.location(tracer.stations[i])
      if econc[i,j] < 0:
        #ax.scatter(loc[1],loc[0],marker='.',edgecolor='k',facecolor='none')
        continue
      if (gconc[i,j] >= 0.01) and (econc[i,j] > 0):
        ax.scatter(loc[1],loc[0],marker='o',edgecolor=colour2,
          facecolor=cmap((np.log10(econc[i,j])+np.abs(np.log10(levels[0])))/(np.log10(levels[-1])+
            np.abs(np.log10(levels[0])))),s=s)

        # ax.scatter(loc[1],loc[0],marker='o',edgecolor=colour2,
        #   facecolor=cmap((np.log10(gconc[i,j])+np.abs(np.log10(levels[0])))/(np.log10(levels[-1])+
        #     np.abs(np.log10(levels[0])))),s=s/2)

      elif econc[i,j] > 0:
        ax.scatter(loc[1],loc[0],marker='^',edgecolor=colour,
          facecolor=cmap((np.log10(econc[i,j])+np.abs(np.log10(levels[0])))/(np.log10(levels[-1])+
            np.abs(np.log10(levels[0])))),s=s)
      elif gconc[i,j] >= 0.01:
        ax.scatter(loc[1],loc[0],marker='x',edgecolor=colour,facecolor=colour,s=s)
      else:
        ax.scatter(loc[1],loc[0],marker='.',edgecolor=colour2,facecolor='none',s=s)

    if captex:
      ax.set_extent( [-85,-70,38,47] )
    else:
      ax.set_extent( [-4.5,25,43.25,66.25] )

    #line1 = Line2D([0],[0],label='Station',  color='w',marker='.',markeredgecolor='k',markerfacecolor='none')
    line2 = Line2D([0],[0],label='Model+Measurement',  color='w',marker='o',markeredgecolor=colour2,markerfacecolor='none',markersize=s/10)
    line25 = Line2D([0],[0],label='Measurement',  color='w',marker='^',markeredgecolor=colour,markerfacecolor='none',markersize=s/10)
    line3 = Line2D([0],[0],label='Model', color='w',marker='x',markeredgecolor=colour,markerfacecolor=colour,markersize=s/10)
    ax.legend(handles=[line2,line25,line3],frameon=False,loc='upper left',fontsize=22)

    if captex:
      cbt=ax.contourf(lonnew, latnew, np.sum(grid.data[2*j:2*j+2,0,:,:],axis=0),cmap=cmap,levels=levels,
        norm=LogNorm(),extend='both',zorder=0)
    else:
      cbt=ax.contour(lonnew, latnew, grid.data[j,0,:,:],cmap=cmap,levels=levels,
        norm=LogNorm(),extend='both',zorder=0,linewidth=0.1)
      cbt=ax.contourf(lonnew, latnew, grid.data[j,0,:,:],cmap=cmap,levels=levels,
        norm=LogNorm(),extend='both',zorder=0)

    cbt.cmap.set_under('white')
    cbt.cmap.set_over(cmap(1.))

    cb=plt.colorbar(cbt,orientation='vertical',ticks=[0.01,0.03,0.1,0.3,1,10],fraction=0.1)
    
    ax.scatter(-2.00833011,48.0583,facecolor="None", edgecolor='r',marker="*",s=100,)

    cb.cmap.set_under('white')
    cb.cmap.set_over(cmap(1.))
    cb.set_label('concentration [ng m$^{-3}$]',fontsize=22)
    ww = np.where(econc[:,j]>=0)[0]
    if len(ww)>0:
      pp=stats.pearsonr(econc[ww,j],gconc[ww,j])
      pp1=stats.spearmanr(econc[ww,j],gconc[ww,j])
    else:
      pp1=[0,0]
    if captex:
      ax.set_title('%i hours, Spearman coefficient: %1.2f' %(grid._data['time'][j]/3600.*2,pp1[0]))
    else:
      ax.set_title('%i hours, Spearman coefficient: %1.2f' %(grid._data['time'][j]/3600.,pp1[0]),fontsize=22)

    ax.set_xticks([0, 10,20], crs=ccrs.PlateCarree())
    ax.set_xticklabels(labels=['0$^{\\circ}$ E','10$^{\\circ}$ E','20$^{\\circ}$ E'])
    ax.set_yticks([50, 60], crs=ccrs.PlateCarree())
    ax.set_yticklabels(labels=['50$^{\\circ}$ N','60$^{\\circ}$ N'])
    #plt.savefig('/home/lucie/CAPTEX/plots/release1/etalog/'+fbasename+'%i.png' %j)
    plt.tight_layout()
    plt.savefig('./etex_%i.png' %j)
    plt.savefig('./etex_%i.pdf' %j)
    plt.close()

